/* This is free and unencumbered software released into the public domain.
 * 
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * See UNLICENSE for complete information.
 *
 * This FFT-code is part of the implementation which implements 
 * our modified Gauss-sieve algorithm specifically 
 * aimed at finding short vectors in an ideal lattice. Ideal lattices are 
 * lattices with additional structure, namely they are also ideals in a ring R.
 * We restrict to lattices which correspond to ideals in the ring of integers 
 * of a cyclotomic number field, namely R = Z[X]/(X^(2^7)+1). We use this structure
 * by computing 128 scalar products by a single ring product using Nussbaumer's
 * symbolic FFT approach (negacyclic convolution). 
 *
 * See for a detailed description of the methods used:
 * Joppe W. Bos, Michael Naehrig, and Joop van de Pol:
 * Sieving for Shortest Vectors in Ideal Lattices:
 * a Practical Perspective. Cryptology ePrint Archive, 
 * Report 2014/880, IACR, 2014.
 * http://eprint.iacr.org/2014/880.pdf
 */

/* Nussbaumer approach, see:
 * H. J. Nussbaumer. Fast polynomial transform algorithms for digital 
 * convolution. Acoustics, Speech and Signal Processing, 
 * IEEE Transactions on, 28(2):205{215, 1980.
 * We followed the description from Knuth:
 * D. E. Knuth. Seminumerical Algorithms. 
 * The Art of Computer Programming. Addison-Wesley, Reading,
 * Massachusetts, USA, 3rd edition, 1997
 * Exercise Exercise 4.6.4.59.
 */

#ifndef __FFT
#define __FFT

#define DATATYPE int32_t

#include "shared.h"

#define add(c,a,b) (c)=(a)+(b)
#define sub(c,a,b) (c)=(a)-(b)
#define mul(c,a,b) (c)=(a)*(b)
#define div2(c,a)  (c)=(a)/2
#define neg(c,a)   (c)=-(a)
#define sqr(c,a)   (c)=(a)*(a)
#define set(c,a)   (c)=(a)

/* Memory management routines */
void FFT_init (void);
void FFT_free (void);

/* Conversions from list vector to FFT and from FFT to inner product vector*/
void toFFT (DATATYPE ***x2, INT *x);
void fromFFT(DATATYPE *x, DATATYPE ***x2, __m128i **X);

/* Convert FFT to the two different representations 1 and 2*/
void FFT2SSE1(__m128i ***fft, DATATYPE ***x2);
void FFT2SSE2(__m128i ***fft, DATATYPE ***x2);

/* Compute ring product. 
 *	x2 is in (reverse) representation 1
 *	y2 is in (forward) representation 2
 */
void FFT_mul(__m128i **Z, __m128i ***x2, __m128i ***y2);

#endif /* __FFT */

