/* This is free and unencumbered software released into the public domain.
 * 
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * See UNLICENSE for complete information.
 *
 * This code implements our modified Gauss-sieve algorithm.
 *
 * See for a detailed description of the methods used:
 * Joppe W. Bos, Michael Naehrig, and Joop van de Pol:
 * Sieving for Shortest Vectors in Ideal Lattices:
 * a Practical Perspective. Cryptology ePrint Archive, 
 * Report 2014/880, IACR, 2014.
 * http://eprint.iacr.org/2014/880.pdf
 */

#include "shared.h"

/* Saves the list every STORE rounds or never if 0 */
#define STORE 0

/* Size increments of main list and stack respectively */
#define SIZE (8*1024)
#define SIZE2 1024

/* Starting list size */
#define INIT_SIZE (128)

/* Samples to be used per round per node */
#define BATCH (16)

FILE *rout;

typedef struct {
  __m128i *v;
  IP inprod;
} element_t;

void element_init (element_t *new) {
  posix_memalign ((void *) &(new->v), 16, (N/8) * sizeof (__m128i));
  
  if (new->v == NULL) {
    fprintf (stderr, "Error: element_init out of memory.\n");
    exit (-1);
  }
}

static void copy_element (element_t *c, element_t *a) {
  memcpy (c->v, a->v, (N/8)*sizeof(__m128i));
  c->inprod = a->inprod;
}

static IP innerproduct (__m128i *a, __m128i *b) {
  __m128i c, sum;
  static int64_t t[2] __attribute__ ((aligned (16)));
  int i;
  static int init = 1;
  static __m128i zero;
  __m128i cmp;

  if (init) {
    uint32_t tmp[4];

    tmp[0] = (uint32_t) 0;
    tmp[1] = (uint32_t) 0;
    tmp[2] = (uint32_t) 0;
    tmp[3] = (uint32_t) 0;
    zero = _mm_loadu_si128 ((__m128i *) tmp);

    init = 0;
  }

  c = _mm_madd_epi16 (a[0], b[0]);
  cmp = _mm_cmplt_epi32 (c, zero);
  sum = _mm_add_epi64 (_mm_unpacklo_epi32 (c, cmp), _mm_unpackhi_epi32 (c, cmp));

  for (i=1; i < N/8; i++) {
    c = _mm_madd_epi16 (a[i], b[i]);
    cmp = _mm_cmplt_epi32 (c, zero);
    sum = _mm_add_epi64 (sum, _mm_add_epi64 (_mm_unpacklo_epi32 (c, cmp), _mm_unpackhi_epi32 (c, cmp)));
  }
  _mm_store_si128 ((__m128i *) t, sum);

  return (t[0]+t[1]);
}

static void vec_submul (__m128i *res, INT c, __m128i *a) {
  int i, j;

  if (c > 0) {
    for (i=0; i < c; i++) {
      for (j=0; j < N/8; j++) {  
        res[j] = _mm_sub_epi16 (res[j], a[j]);
      }
    }
  } else {
    for (i=0; i < -c; i++) {
      for (j=0; j < N/8; j++) { 
        res[j] = _mm_add_epi16 (res[j], a[j]);
      }
    }
  }
}

static int reduce (element_t *list, element_t *samp){
  IP inprodnew, inprodnewa; 
  int order = (samp->inprod < list->inprod);

  inprodnew = innerproduct (samp->v, list->v); 
  inprodnewa = (inprodnew < 0) ? -inprodnew : inprodnew; 
  inprodnewa <<= 1; 
  if (order && inprodnewa > samp->inprod) { 
    INT mu; 
    mu = ((inprodnew^samp->inprod)<0); mu = -mu*2+1; 
    vec_submul (list->v, mu, samp->v); 
    UPDATE_MU_STAR(list,samp) 
    //Vector list has been reduced
    return 1;
  } else if (!order && inprodnewa > list->inprod) { 
    INT mu; 
    mu = ((inprodnew^samp->inprod)<0); mu = -mu*2+1; 
    vec_submul (samp->v, mu, list->v); 
    UPDATE_MU_STAR(samp,list)
    //Vector samp has been reduced
    return 2;
  } 
  return 0;
}

int client(int rank, int size){
  MPI_Status status;
  INT **b = NULL; //Basis, used for sampling
  INT *sampv = NULL; //Vector, used for sampling
  element_t *L = NULL; //List
  int64_t *L_bits = NULL; //Mask indicating list occupancy
  int listsize; //Number of vectors currently in the list
  int Lsize, oldLsize; //Current list storage size
  
  element_t *S = NULL; //Stack
  int stacksize; //Number of vectors currently in the stack
  int Ssize; //Current stack storage size
  element_t new;  
  element_t *Q = NULL; //Sample list
  int Qsize = BATCH*size;
  int Qoffset = BATCH*rank;
  int QoffsetB = Qoffset+BATCH;
  //int16_t *com_buffer = NULL; //Used to send vectors
  //int comSize;
  __m128i *com_buffer = NULL; //Used to send vectors
  int comSize;
  int64_t *survivors = NULL, *rsurv = NULL; //Masks indicating survivors
  int64_t *reqsurv = NULL, *reqrecv = NULL; //Masks for requesting vectors
  int *reqnums = NULL;  
  int16_t svp[N];  
  int i, j, k, l;
  int64_t max, minimum, goal;
  double **gs, **mu, *lbj, *vars, t; //Gram-Schmidt used for sampling

  int collisions, iterations;  
  int terminate;  
  double starttime, inittime, mintime, ghtime;
  
  struct { 
    int lsize; 
    int rank; 
  } lrsend, lrrecv;

  struct _inprod_rank {
    long inprod;
    int rank;
  } *irsend = NULL, *irrecv = NULL, minsend, minrecv;
  
  irsend = (struct _inprod_rank*) malloc (Qsize * sizeof(struct _inprod_rank));
  irrecv = (struct _inprod_rank*) malloc (Qsize * sizeof(struct _inprod_rank));
  reqnums = (int32_t *) malloc (size * sizeof(int32_t));
  if (irsend == NULL || irrecv == NULL || reqnums == NULL) {
    fprintf (stderr, "Memory error\n");
    return -1;
  }

  //Randomize seed
  {
    struct timeval time; 
    gettimeofday(&time,NULL);
    srand( ((rank << 16) * getpid()) ^ ((time.tv_sec * 1000) + (time.tv_usec / 1000)) ); 
  }

  element_init(&new);
  
  //Initialize memory
  Lsize = SIZE;
  oldLsize = 0;
  L = (element_t *) malloc (SIZE * sizeof (element_t));
  for(i=0; i < SIZE; i++){
    element_init((L+i));
  }
  L_bits = (int64_t *) malloc ((SIZE/64+1)*sizeof(int64_t));
  
#define INCREMENT_LISTSIZE                                             \
listsize++;                                                            \
if(listsize == Lsize) {                                                \
  oldLsize = Lsize;                                                    \
  Lsize += SIZE;                                                       \
  L = (element_t *) realloc (L, Lsize*sizeof(element_t));              \
  for(k=oldLsize; k < Lsize; k++) element_init(L+k);                   \
  L_bits = (int64_t *) realloc (L_bits, (Lsize/64+1)*sizeof(int64_t)); \
  for (l=oldLsize; l < Lsize; l++) { clearbit(L_bits,l); }             \
}

  Ssize = SIZE2;
  S = (element_t *) malloc (SIZE2 * sizeof (element_t));
  for(i=0; i < SIZE2; i++){
    element_init((S+i));
  }

#define INCREMENT_STACKSIZE                                    \
stacksize++;                                                   \
if(stacksize == Ssize) {                                       \
  l = Ssize;                                                   \
  Ssize += SIZE2;                                              \
  S = (element_t *) realloc (S, (Ssize) * sizeof (element_t)); \
  for(;l<Ssize; l++) element_init((S+l));                      \
}
  Q = (element_t *) malloc (Qsize * sizeof (element_t));
  for(i=0; i < Qsize; i++){
    element_init((Q+i));
  }
  
  //comSize = N*Qsize;
  //com_buffer = (int16_t *) malloc ((N*Qsize) * sizeof (int16_t));
  comSize = (N/8)*SIZE2;
  posix_memalign ((void *) &com_buffer, 16, (comSize) * sizeof (__m128i)); 
  
  survivors = (int64_t *) malloc ((Qsize/64+1)*sizeof(int64_t));
  rsurv = (int64_t *) malloc ((Qsize/64+1)*sizeof(int64_t));
  reqsurv = (int64_t *) malloc((Qsize/64+1) * size * sizeof(int64_t));
  reqrecv = (int64_t *) malloc((Qsize/64+1) * sizeof(int64_t));
  
  for(i = 0; i < Qsize; i++) irsend[i].rank = rank;
  
  //Initialize memory for basis
  b = (INT**) malloc (N * sizeof (INT*));
  for (i=0; i < N; i++) {
    b[i]  = (INT*) malloc (N * sizeof (INT));
  }
  posix_memalign ((void *) &(sampv), 16, (N) * sizeof (INT));
  
  max = readbasis (fileName, b);
  if (max == 0) {
    printf ("Read basis error\n");
    exit (0);
    return -1;
  }
  
  gs  = (double **) malloc (N * sizeof (double*));
  mu  = (double **) malloc (N * sizeof (double*));
  lbj = (double *)  malloc (N * sizeof (double));
  vars = (double *)  malloc (N * sizeof (double));
        
  for (i=0; i < N; i++) {
    gs[i] = (double *)  malloc (N * sizeof (double));
    mu[i] = (double *)  malloc (N * sizeof (double));
  }

  for(i=0; i<N; i++) sampv[i] = 0;
  for(i=0; i<comSize; i += (N/8)) toSSE2(com_buffer+i, sampv);

  gram_schmidt (gs, mu, lbj, b, &goal);
  sample_init(lbj,vars);

  t=(log((double)n)*log((double)n))/70.; // ToDO: find the optimal value for t  
  
  toSSE2(new.v, b[0]);
  
  minimum = innerproduct(new.v,new.v); 
  fromSSE2(svp,new.v);
   
  //Populate the list with the first element 
  if(rank < n){
    toSSE2(L[0].v,b[rank]);
  } else {
    sample (sampv,b,mu,vars,t);
    toSSE2(L[0].v,sampv);
  }
  setbit(L_bits,0);
  L[0].inprod = innerproduct(L[0].v,L[0].v); 
  listsize = 1;

  stacksize = 0;
  for(i=rank+size; i < n; i+=size){
    toSSE2(S[stacksize].v,b[i]);
    S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
    stacksize++;
  }
  
  for(;stacksize < INIT_SIZE-1; stacksize++){
    sample (sampv,b,mu,vars,t);
    toSSE2(S[stacksize].v,sampv);
    S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
  }
  
  k=0; 
  starttime =  0.0;
  inittime  =  0.0;
  mintime   =  0.0;
  ghtime    = -1.0;
  
  if(rank == 0) fprintf(rout, "Start minimum %ld (%ld)\n",minimum,goal);

  //Perform reduction until stack is empty
  collisions = 0;
  terminate = 0;
  starttime = MPI_Wtime();
  while(stacksize > 0){ 
    int firstempty = Lsize;
    j=0;
    copy_element(&new, (S+stacksize-1));
    stacksize--;
    
    for(i = 0; i < Lsize; i++){
      if(getbit(L_bits, i) != 0){
        int red = reduce(&new, (L+i));
        j++;
        
        if (red & 1) { 
          if (new.inprod != 0) { 
            i=-1; 
            j=0; 
            continue;
          } else { 
            collisions++; 
            break; 
          } 
        } else if (red & 2) { 
          listsize--; 
          j--; 
          clearbit (L_bits, i); 
          if (L[i].inprod == 0) {
            collisions++; 
          } else { 
            copy_element((S+stacksize),(L+i));
            INCREMENT_STACKSIZE
          }
        }
      } else if(i < firstempty) {
        firstempty = i;
      }
      
      if (j == listsize && (firstempty < Lsize)) break;
    }
    if (new.inprod == 0){
      continue;
    } else if (new.inprod < minimum){
      minimum = new.inprod;
      fprintf(rout, "Rank %d: New minimum %lu\n", rank, minimum);
      fromSSE2(svp,L[i].v);
      if (minimum < goal) {
        if (ghtime < 0) ghtime = MPI_Wtime()-starttime;
        fprintf(rout,"Goal reached! Entering termination.\n");
        terminate = 1;
      }
    }
    if (firstempty == Lsize){
      //Error
      fprintf (stderr, "Error: list error.\n");
      exit(1);
    }
    copy_element((L+firstempty), &new);
    setbit(L_bits,firstempty);
    INCREMENT_LISTSIZE
  }
   
  MPI_Barrier (MPI_COMM_WORLD); //Everyone has a list of size INIT_SIZE (minus collisions?)
  inittime = MPI_Wtime() - starttime;
  if(rank == 0) fprintf(rout, "Barrier passed. (%.2lf)\n",inittime);
  
  iterations = 0;
  while(1) { //The main loop
    iterations++;
    
    //Root samples all vectors
    if(rank == 0) {
      j=0;
      while(stacksize > 0 && j < Qsize) {
        copy_element((Q+j), (S+stacksize-1));
        fromSSE2(sampv,Q[j].v);
        stacksize--; 
        memcpy ((com_buffer+j*(N/8)), Q[j].v, (N/8)*sizeof(__m128i));
        j++;
      }
      for(; j < Qsize; j++) {
        sample(sampv, b,mu,vars,t);
        toSSE2(Q[j].v,sampv);
        Q[j].inprod = innerproduct(Q[j].v,Q[j].v);
        memcpy ((com_buffer+j*(N/8)), Q[j].v, (N/8)*sizeof(__m128i));
      }
    }
    
    MPI_Bcast (com_buffer,(2*N*Qsize),MPI_BYTE,0,MPI_COMM_WORLD); //Samples are broadcast
    
    if(rank > 0) {
      for(i=0;i<Qsize;i++){
        memcpy (Q[i].v, com_buffer+i*(N/8), (N/8)*sizeof(__m128i));
        Q[i].inprod = innerproduct(Q[i].v,Q[i].v);
      }
    }
    
    //Prep survivor mask
    for(l=0; l < Qsize; l++) {
      setbit(survivors, l);
    }
    for(l = Qsize; l < 64*(Qsize/64 + 1); l++) {
      clearbit(survivors, l);
    }
    
    //Reduce all samples <-> list.
    for(i = 0; i < Lsize; i++) {
      j = 0;
      if(getbit(L_bits, i) != 0) {
        j++;
        //Take L[i] and <-> it with the samples, no insertions
              
        for(k = 0; k < Qsize; k++) {
          int red = reduce((L+i),(Q+k));
          
          if (red & 1) { 
            clearbit (L_bits, i); 
            if (L[i].inprod == 0) { 
              collisions++; 
            } else {
              //Save on local stack, send to root at end
              copy_element((S+stacksize),(L+i));
              INCREMENT_STACKSIZE
            }
            listsize--; j--; 
            break;
          } else if (red & 2) { 
            clearbit(survivors, k);
            if(Q[k].inprod == 0) collisions++;
          }
        }
      }
      if(j == listsize) break;
    }

    //Reduce samples <-> your samples
#define SAMP2SAMP(k,i)                  \
if(getbit(survivors, i) != 0){          \
  int red = reduce((Q+k),(Q+i));        \
  if (red & 1) {                        \
    clearbit(survivors, k);             \
    if (Q[k].inprod == 0) collisions++; \
    break;                              \
  } else if (red & 2) {                 \
    clearbit (survivors, i);            \
    if (Q[i].inprod == 0) collisions++; \
  }                                     \
}
    /*Sample at position k is reduced 
    wrt all samples of same parity before 
    it and all samples of differing parity after it*/
    for(k = Qoffset; k < QoffsetB; k++){
      if(getbit(survivors,k) != 0){
        for(i = k%2; i < Qoffset; i += 2){
          SAMP2SAMP(k,i);
        }
        if(getbit(survivors,k) == 0) continue;
        for(i=k+1; i < QoffsetB; i++){
          SAMP2SAMP(k,i);
        }
        if(getbit(survivors,k) == 0) continue;
        for(i=QoffsetB+((QoffsetB-k)%2); i < Qsize; i+=2){
          SAMP2SAMP(k,i);
        }
      }
    }
#undef SAMP2SAMP

    //Collectively figure out who survived
    MPI_Allreduce (survivors,rsurv,8*(Qsize/64+1),MPI_BYTE,MPI_BAND,MPI_COMM_WORLD); 
    
    lrsend.lsize = listsize;
    lrsend.rank = rank;
    
    if(terminate > 0 || collisions > 0.1 * (size*listsize) + 200){
      lrsend.lsize = -1;
    }
    
    //Collectively compute who has the smallest list (tiebreaker being smaller rank)
    MPI_Allreduce (&lrsend,&lrrecv,1,MPI_2INT,MPI_MINLOC,MPI_COMM_WORLD); 
    
    if(lrrecv.lsize < 0) break;
    
    if(rank == lrrecv.rank) { //Insert the survivors into your list
      i=0;
      for(k=0; k < Qsize; k++) {
        if(getbit(rsurv,k) != 0) {
          for(; i < Lsize; i++) {
            if(getbit(L_bits,i) == 0) {
              setbit(L_bits,i);
              copy_element((L+i),(Q+k));
              if(L[i].inprod < minimum) {
                minimum = L[i].inprod;
                mintime = MPI_Wtime() - starttime;
                fprintf(rout,"Rank %d: New minimum %lu (%.2lf)\n", rank, minimum, mintime);
                fromSSE2(svp,L[i].v);
                if(minimum < goal) {
                  if(ghtime < 0) ghtime = MPI_Wtime()-starttime;
                  fprintf(rout,"Goal reached! Entering termination.\n");
                  terminate = 1;
                }
              }
              INCREMENT_LISTSIZE
              break;
            }
          }
        }
      }
    }
    
    fflush(stdout);
    MPI_Barrier(MPI_COMM_WORLD);

    for(i=0; i < Qsize; i++) {
      irsend[i].inprod = Q[i].inprod;
      irsend[i].rank = rank;
    }
    
    //Root computes who has smallest innerproduct for non-survivors.
    MPI_Reduce(irsend,irrecv,Qsize,MPI_LONG_INT,MPI_MINLOC,0,MPI_COMM_WORLD);
    
    if(rank == 0) { //Root prepares mask for which vectors to request from nodes
      for(i=0; i < size; i++) {
        reqnums[i] = 0;
        for(j=0;j < (Qsize/64+1); j++) reqsurv[i*(Qsize/64+1)+j] = 0;
      }
      
      for(i=0;i<Qsize;i++) {
        if(getbit(rsurv,i) == 0 && irrecv[i].inprod > 0) {
          j = irrecv[i].rank;
          reqnums[j]++;
          setbit(reqsurv,j*(64)*(Qsize/64+1)+i);
        }
      }
    }
    //Root sends mask for which vectors to request from each node
    MPI_Scatter (reqsurv,8*(Qsize/64+1),MPI_BYTE,reqrecv,8*(Qsize/64+1),MPI_BYTE,0,MPI_COMM_WORLD);
    
    //Consolidate all non-surviving samples in the root-stack
    if(rank == 0) { //Root receives everyone's 
      for(i = 0; i < Qsize; i++){
        if(getbit(reqrecv,i) != 0){
          copy_element((S+stacksize),(Q+i));
          INCREMENT_STACKSIZE
        }
      }
      for(i = 1; i < size; i++) {
        if(reqnums[i] > 0){
          MPI_Recv(com_buffer,2*(N*reqnums[i]),MPI_BYTE,i,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
          for(j=0; j < reqnums[i]; j++){
            memcpy (S[stacksize].v, com_buffer+j*(N/8), (N/8)*sizeof(__m128i));
            S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v); //TODO: Can this be helped?
            INCREMENT_STACKSIZE
          }
        }
      }
      
    } else {
      j=0;
      for(i=0; i < Qsize; i++) {
        if(getbit(reqrecv,i) != 0) {
          memcpy (com_buffer+j*(N/8), Q[i].v, (N/8)*sizeof(__m128i));
          j++;
        }
      }
      if(j>0) MPI_Send (com_buffer,2*(N*j),MPI_BYTE,0,0,MPI_COMM_WORLD);
    }
    
    //Root gathers everyone's stacksizes
    MPI_Gather (&stacksize,1,MPI_INT,reqnums,1,MPI_INT,0,MPI_COMM_WORLD);
        
    if(rank == 0){ //Root receives everyone's stack contents
      for(i=1; i < size; i++){
        int incstacksize = reqnums[i];
        if(incstacksize > 0) {
          if((N/8)*incstacksize > comSize){
            __m128i *tmp; 
            int newcomSize = (N/8)*incstacksize;            
            posix_memalign ((void *) &tmp, 16, (newcomSize) * sizeof (__m128i));
            free(com_buffer);
            com_buffer = tmp;
            comSize = newcomSize;
          }
          MPI_Recv(com_buffer,2*(N*incstacksize),MPI_BYTE,i,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
          for(k=0; k < incstacksize; k++) {
            memcpy (S[stacksize].v, com_buffer+k*(N/8), (N/8)*sizeof(__m128i));
            S[stacksize].inprod = innerproduct(S[stacksize].v,S[stacksize].v);
            INCREMENT_STACKSIZE
          }
        }
      }
    } else if (stacksize>0) { //Everyone else sends their stack to the root
      if((N/8)*stacksize > comSize){
        __m128i *tmp; 
        int newcomSize = (N/8)*stacksize;
        posix_memalign ((void *) &tmp, 16, (newcomSize) * sizeof (__m128i));
        free(com_buffer);
        com_buffer = tmp;
        comSize = newcomSize;
      }
      for(k=0; k < stacksize; k++) {
        memcpy (com_buffer+k*(N/8), S[k].v, (N/8)*sizeof(__m128i));
      }
      MPI_Send (com_buffer,(2*N*stacksize),MPI_BYTE,0,0,MPI_COMM_WORLD);
      stacksize = 0;
    }
        
    if(rank == 0 && iterations%1024 == 0) { fprintf(rout, "R = %d, L = %d, C = %d, S = %d (%2.lf)\n", iterations, listsize, collisions, stacksize, MPI_Wtime() - starttime); fflush(rout); }
  }
  
  {
    int tlist, tstack, tcoll;
    double minght;
    if(ghtime < 0 || ghtime == 0.0) ghtime = MPI_Wtime() - starttime;
    MPI_Reduce (&listsize,&tlist,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce (&stacksize,&tstack,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce (&collisions,&tcoll,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
    MPI_Reduce (&ghtime,&minght,1,MPI_DOUBLE,MPI_MIN,0,MPI_COMM_WORLD);
    if(rank == 0) { fprintf(rout, "R = %d, TL = %d, TC = %d, TS = %d (T %2.lf, GHT %2.lf)\n", iterations, tlist, tcoll, tstack, MPI_Wtime() - starttime, minght); fflush(rout); }
    minsend.inprod = minimum;
    minsend.rank = rank;
    MPI_Allreduce(&minsend, &minrecv, 1, MPI_LONG_INT,MPI_MINLOC, MPI_COMM_WORLD);
    if(rank == minrecv.rank) {
      FILE *file;
      file = fopen("Min.out", "w");
      fprintf(file, "%lu (%.2lf)\n", minimum, mintime);
      fprintf(file, "[");
      for(l=0;l<N;l++) fprintf(file, "%d ", svp[l]);
      fprintf(file,"]\n"); fflush(file);
    }
    
    fprintf(rout, "%lu (%.2lf, %.2lf)\n", minimum, mintime,ghtime);
    fprintf(rout, "[");
    for(l=0;l<n;l++) fprintf(rout, "%d ", svp[l]);
    fprintf(rout,"]\n"); fflush(rout);
    
  }
  return 1;
}

int main (int argc, char* argv[]) {
  int rank, size;

  if (argc != 3) {
    printf ("Usage %s dim filename\n", argv[0]);
    return -1;
  }
  n = atoi (argv[1]);
  N = 8*ceil(n/8.0);
  
  fileName = (char *) malloc (64 * sizeof(char));
  fileName = argv[2];

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  
  if(rank == 0) rout = fopen("Root.out", "w");
  else{
    char fName[32];
    sprintf(fName, "Client%d.out", rank);
    rout = fopen(fName, "w");
  }
  client (rank,size);

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
  return 0;
}

