/* This is free and unencumbered software released into the public domain.
 * 
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * See UNLICENSE for complete information.
 *
 * This code implements our modified Gauss-sieve algorithm.
 * This file contains shared routines between the different implementations.
 *
 * See for a detailed description of the methods used:
 * Joppe W. Bos, Michael Naehrig, and Joop van de Pol:
 * Sieving for Shortest Vectors in Ideal Lattices:
 * a Practical Perspective. Cryptology ePrint Archive, 
 * Report 2014/880, IACR, 2014.
 * http://eprint.iacr.org/2014/880.pdf
 */

#include "shared.h"

uint64_t readbasis (char *str, INT **b) {
  int64_t i, j, k;
  char num[128], c = '\0';
  FILE *in = NULL;
  uint64_t max;
  
  max = 0;

  in = fopen (str, "r");
  if (in == NULL) return 0;

  for (i=0; i < n; i++) {
    for (j=0; j < n; j++) {
      c = fgetc (in);
      while (c == ' ' || c == '[' || c == ']' || c == '\n' || c == '\r') c = fgetc (in);      
      k=0;
      while (c != ' ' && c != ']') {        
        num[k++] = c;
        c = fgetc (in);
      }
      num[k] = '\0';
      b[i][j] = atoi (num);
      if (abs(b[i][j]) > max) max = abs(b[i][j]);
    }
    for(j = n; j < N; j++){
      b[i][j] = 0;
    }
  }
  while (c == ' ' || c == '[' || c == ']' || c == '\n' || c == '\r') c = fgetc (in);
  
  fclose (in);
  return max;
}

void toSSE2(__m128i *c, INT *a){
  int i;
  for (i=0; i < N; i+=8) { 
    c[i/8] = _mm_load_si128 ((__m128i *) (a+i));
  }
}

void fromSSE2(INT *c, __m128i *a){
  int i;
  for (i=0; i < N/8; i++) {
    _mm_store_si128 ((__m128i *) (c+8*i), a[i]);
  }
}

static void copy_double_int (double *a, INT *b, int l) {
  int i;
  for (i=0; i < l; i++) {
    a[i] = (double) b[i];
  }
}

static void vec_addmul_double (double *res, double c, double *a, int l) {
  int i;
  for (i=0; i < l; i++) {
    res[i] += (c * a[i]);
  }
}

static IP innerproduct_int (INT *v, INT *w, int l) {
  int i;
  IP sum = 0;
  for (i=0; i < l; i++) {
        sum += (IP) v[i] * w[i];
  }
  return sum;
}


/* Code to compute Gram-Schmidt, the technique is from
 *
 * Phong Q. Nguyen, Damien Stehlé. 
 * "An LLL algorithm with quadratic complexity." 
 * In SIAM Journal on Computing 39.3, pp. 874-903. SIAM, 2009.
 */
void gram_schmidt (double **gs, double **mu, double *lbj, INT **b, int64_t *goal) {
  int i, j, k;
  IP gram[n][n];
  double r[n][n];
  double tmp1, tmp2;

  for(i=0; i < n; i++){
    for(j=0; j < i; j++){
      gram[i][j] = innerproduct_int(b[i],b[j],n);
      gram[j][i] = gram[i][j];
    }
    gram[i][i] = innerproduct_int(b[i],b[i],n);
  }
  
  tmp2 = 0.0;
  for(i=0; i < n; i++){
    copy_double_int (gs[i], b[i], n);
    for(j = 0; j <= i; j++){
      tmp1 =gram[i][j];
      
      for (k = 0; k < j; k++) {
        tmp1 -= mu[j][k] * r[i][k];
      }
      r[i][j] = tmp1;
      if (i > j) {
        mu[i][j] = tmp1/r[j][j];
        vec_addmul_double (gs[i], -mu[i][j], gs[j], n);
      } else {
        lbj[i] = r[i][i];
      }
    }
    tmp2 += log(r[i][i]);
  }
  
  tmp2+= 2.0*log(tgamma(n/2.0 + 1));
  tmp2/=(double)n;
  *goal = (int64_t) ceil(1.05*1.05/M_PI*exp(tmp2));
}

/* Code to sample from Discrete Gaussians, the technique is from
 *
 * Craig Gentry, Chris Peikert, and Vinod Vaikuntanathan. 
 * "Trapdoors for hard lattices and new cryptographic constructions." 
 * In Proceedings of the fortieth annual ACM symposium on Theory of computing, pp. 197-206. ACM, 2008.
 */

/* TODO: Possibly replace with better random function. */
#define RANDOM(x) x = (uint32_t) (((uint64_t)rand()<<(uint64_t)15) ^ ((uint64_t)rand()<<(uint64_t)30) ^ (uint64_t)rand())

#define tf(x) log((double) (x))

void sample_init (double* lbj, double* vars) {
  int i;
  double max = 0.0, var;

  for (i=0; i < n; i++) {
    if(lbj[i] > max) max = lbj[i];
  }
  var = max;
  for(i = 0; i < n; i++) {
    vars[i] = var/lbj[i];
  }
}

int32_t sampleZ(double c, double var, double t) {
  int32_t lower,upper,x;
  double rho,tmp;
  uint32_t r;
  double sd = sqrt(var) * t;
        
  lower = (int32_t) ceil (c-sd);
  upper = (int32_t) floor (c+sd);
  while(1){
    RANDOM(r);
    x = lower + (int32_t)(ceil((upper-lower)*(double)r/(double)UINT_MAX - 0.5));
    tmp = (double)x - c;
    rho = exp(-M_PI*tmp*tmp/var);
    RANDOM(r);
    if((double)r/(double)UINT_MAX <= rho) return x;
  }
}

void sample (INT *v, INT **b, double **mu, double *vars, double t) {
  static int init=1;
  static int32_t *x;
  int i,j, iszero;
  double center;
  /*INT v[N];
  __m128i *V;
  V = new->v;*/

  if (init) {
    init=0;
    // pittfall: this is the same for all n in subsequent calls
    x = (int32_t *) malloc (n * sizeof (int32_t));
  }

  do {
    iszero = 0;
    x[n-1] = sampleZ(0.0,vars[n-1], t);
   
    for(i = n-2; i>=0; i--){
      center = 0.0;
      for (j=i+1; j < n; j++) {
        center -= (mu[j][i] * x[j]);
      }
      x[i] = sampleZ(center,vars[i],t);   
    }
    for(i=0; i<N; i++) {
      v[i] = 0;
    }
    for (i=0; i < N; i++) {
      for (j=0; j < n; j++) {
        v[i] += (x[j] * b[j][i]);
      }
      iszero += (v[i] == 0);
    }
  } while (iszero == N);
  
  //toSSE2(V,v);
  //new->inprod = innerproduct(new->v,new->v);
}
